import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.histogram_hs01 import deserialise_hs01, serialise_hs01


def create_test_data_with_type(numpy_type):
    return {
        "source": "some_source",
        "timestamp": 123456,
        "current_shape": [5],
        "dim_metadata": [
            {
                "length": 5,
                "unit": "m",
                "label": "some_label",
                "bin_boundaries": np.array([0, 1, 2, 3, 4, 5]).astype(numpy_type),
            }
        ],
        "last_metadata_timestamp": 123456,
        "data": np.array([1, 2, 3, 4, 5]).astype(numpy_type),
        "errors": np.array([5, 4, 3, 2, 1]).astype(numpy_type),
        "info": "info_string",
    }


class TestSerialisationHs01:
    def _check_metadata_for_one_dimension(self, data, original_data):
        assert np.array_equal(data["bin_boundaries"], original_data["bin_boundaries"])
        assert data["length"] == original_data["length"]
        assert data["unit"] == original_data["unit"]
        assert data["label"] == original_data["label"]

    def test_serialises_and_deserialises_hs01_message_correctly_for_full_1d_data(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_hist = {
            "source": "some_source",
            "timestamp": 123456,
            "current_shape": [5],
            "dim_metadata": [
                {
                    "length": 5,
                    "unit": "m",
                    "label": "some_label",
                    "bin_boundaries": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]),
                }
            ],
            "last_metadata_timestamp": 123456,
            "data": np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
            "errors": np.array([5.0, 4.0, 3.0, 2.0, 1.0]),
            "info": "info_string",
        }

        buf = serialise_hs01(original_hist)
        hist = deserialise_hs01(buf)

        assert hist["source"] == original_hist["source"]
        assert hist["timestamp"] == original_hist["timestamp"]
        assert hist["current_shape"] == original_hist["current_shape"]
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][0], original_hist["dim_metadata"][0]
        )
        assert np.array_equal(hist["data"], original_hist["data"])
        assert np.array_equal(hist["errors"], original_hist["errors"])
        assert hist["info"] == original_hist["info"]
        assert (
            hist["last_metadata_timestamp"] == original_hist["last_metadata_timestamp"]
        )

    def test_serialises_and_deserialises_hs01_message_correctly_for_minimal_1d_data(
        self,
    ):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_hist = {
            "timestamp": 123456,
            "current_shape": [5],
            "dim_metadata": [
                {
                    "length": 5,
                    "unit": "m",
                    "label": "some_label",
                    "bin_boundaries": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]),
                }
            ],
            "data": np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
        }
        buf = serialise_hs01(original_hist)

        hist = deserialise_hs01(buf)
        assert hist["source"] == ""
        assert hist["timestamp"] == original_hist["timestamp"]
        assert hist["current_shape"] == original_hist["current_shape"]
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][0], original_hist["dim_metadata"][0]
        )
        assert np.array_equal(hist["data"], original_hist["data"])
        assert len(hist["errors"]) == 0
        assert hist["info"] == ""

    def test_serialises_and_deserialises_hs01_message_correctly_for_full_2d_data(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_hist = {
            "source": "some_source",
            "timestamp": 123456,
            "current_shape": [2, 5],
            "dim_metadata": [
                {
                    "length": 2,
                    "unit": "b",
                    "label": "y",
                    "bin_boundaries": np.array([10.0, 11.0, 12.0]),
                },
                {
                    "length": 5,
                    "unit": "m",
                    "label": "x",
                    "bin_boundaries": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]),
                },
            ],
            "last_metadata_timestamp": 123456,
            "data": np.array([[1.0, 2.0, 3.0, 4.0, 5.0], [6.0, 7.0, 8.0, 9.0, 10.0]]),
            "errors": np.array([[5.0, 4.0, 3.0, 2.0, 1.0], [10.0, 9.0, 8.0, 7.0, 6.0]]),
            "info": "info_string",
        }
        buf = serialise_hs01(original_hist)

        hist = deserialise_hs01(buf)
        assert hist["source"] == original_hist["source"]
        assert hist["timestamp"] == original_hist["timestamp"]
        assert hist["current_shape"] == original_hist["current_shape"]
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][0], original_hist["dim_metadata"][0]
        )
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][1], original_hist["dim_metadata"][1]
        )
        assert np.array_equal(hist["data"], original_hist["data"])
        assert np.array_equal(hist["errors"], original_hist["errors"])
        assert hist["info"] == original_hist["info"]
        assert (
            hist["last_metadata_timestamp"] == original_hist["last_metadata_timestamp"]
        )

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_hist = {
            "timestamp": 123456,
            "current_shape": [5],
            "dim_metadata": [
                {
                    "length": 5,
                    "unit": "m",
                    "label": "some_label",
                    "bin_boundaries": np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0]),
                }
            ],
            "data": np.array([1.0, 2.0, 3.0, 4.0, 5.0]),
        }
        buf = serialise_hs01(original_hist)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_hs01(buf)

    def test_serialises_and_deserialises_hs01_message_correctly_for_int_array_data(
        self,
    ):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_hist = {
            "source": "some_source",
            "timestamp": 123456,
            "current_shape": [5],
            "dim_metadata": [
                {
                    "length": 5,
                    "unit": "m",
                    "label": "some_label",
                    "bin_boundaries": np.array([0, 1, 2, 3, 4, 5]),
                }
            ],
            "last_metadata_timestamp": 123456,
            "data": np.array([1, 2, 3, 4, 5]),
            "errors": np.array([5, 4, 3, 2, 1]),
            "info": "info_string",
        }

        buf = serialise_hs01(original_hist)
        hist = deserialise_hs01(buf)

        assert hist["source"] == original_hist["source"]
        assert hist["timestamp"] == original_hist["timestamp"]
        assert hist["current_shape"] == original_hist["current_shape"]
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][0], original_hist["dim_metadata"][0]
        )
        assert np.array_equal(hist["data"], original_hist["data"])
        assert np.array_equal(hist["errors"], original_hist["errors"])
        assert hist["info"] == original_hist["info"]
        assert (
            hist["last_metadata_timestamp"] == original_hist["last_metadata_timestamp"]
        )

    def test_serialise_and_deserialise_hs01_message_returns_int32_type(self):
        original_hist = create_test_data_with_type(np.int32)

        buf = serialise_hs01(original_hist)
        hist = deserialise_hs01(buf)

        assert np.issubdtype(
            hist["dim_metadata"][0]["bin_boundaries"].dtype,
            original_hist["dim_metadata"][0]["bin_boundaries"].dtype,
        )
        assert np.issubdtype(hist["data"].dtype, original_hist["data"].dtype)
        assert np.issubdtype(hist["errors"].dtype, original_hist["errors"].dtype)

    def test_serialise_and_deserialise_hs01_message_returns_int64_type(self):
        original_hist = create_test_data_with_type(np.int64)

        buf = serialise_hs01(original_hist)
        hist = deserialise_hs01(buf)

        assert np.issubdtype(
            hist["dim_metadata"][0]["bin_boundaries"].dtype,
            original_hist["dim_metadata"][0]["bin_boundaries"].dtype,
        )
        assert np.issubdtype(hist["data"].dtype, original_hist["data"].dtype)
        assert np.issubdtype(hist["errors"].dtype, original_hist["errors"].dtype)

    def test_serialise_and_deserialise_hs01_message_returns_float32_type(self):
        original_hist = create_test_data_with_type(np.float32)

        buf = serialise_hs01(original_hist)
        hist = deserialise_hs01(buf)

        assert np.issubdtype(
            hist["dim_metadata"][0]["bin_boundaries"].dtype,
            original_hist["dim_metadata"][0]["bin_boundaries"].dtype,
        )
        assert np.issubdtype(hist["data"].dtype, original_hist["data"].dtype)
        assert np.issubdtype(hist["errors"].dtype, original_hist["errors"].dtype)

    def test_serialise_and_deserialise_hs01_message_returns_float64_type(self):
        original_hist = create_test_data_with_type(np.float64)

        buf = serialise_hs01(original_hist)
        hist = deserialise_hs01(buf)

        assert np.issubdtype(
            hist["dim_metadata"][0]["bin_boundaries"].dtype,
            original_hist["dim_metadata"][0]["bin_boundaries"].dtype,
        )
        assert np.issubdtype(hist["data"].dtype, original_hist["data"].dtype)
        assert np.issubdtype(hist["errors"].dtype, original_hist["errors"].dtype)

    def test_serialises_and_deserialises_hs01_message_correctly_when_float_input_is_not_ndarray(
        self,
    ):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_hist = {
            "source": "some_source",
            "timestamp": 123456,
            "current_shape": [2, 5],
            "dim_metadata": [
                {
                    "length": 2,
                    "unit": "b",
                    "label": "y",
                    "bin_boundaries": [10.0, 11.0, 12.0],
                },
                {
                    "length": 5,
                    "unit": "m",
                    "label": "x",
                    "bin_boundaries": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
                },
            ],
            "last_metadata_timestamp": 123456,
            "data": [[1.0, 2.0, 3.0, 4.0, 5.0], [6.0, 7.0, 8.0, 9.0, 10.0]],
            "errors": [[5.0, 4.0, 3.0, 2.0, 1.0], [10.0, 9.0, 8.0, 7.0, 6.0]],
            "info": "info_string",
        }
        buf = serialise_hs01(original_hist)

        hist = deserialise_hs01(buf)
        assert hist["source"] == original_hist["source"]
        assert hist["timestamp"] == original_hist["timestamp"]
        assert hist["current_shape"] == original_hist["current_shape"]
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][0], original_hist["dim_metadata"][0]
        )
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][1], original_hist["dim_metadata"][1]
        )
        assert np.array_equal(hist["data"], original_hist["data"])
        assert np.array_equal(hist["errors"], original_hist["errors"])
        assert hist["info"] == original_hist["info"]
        assert (
            hist["last_metadata_timestamp"] == original_hist["last_metadata_timestamp"]
        )

    def test_serialises_and_deserialises_hs01_message_correctly_when_int_input_is_not_ndarray(
        self,
    ):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_hist = {
            "source": "some_source",
            "timestamp": 123456,
            "current_shape": [2, 5],
            "dim_metadata": [
                {
                    "length": 2,
                    "unit": "b",
                    "label": "y",
                    "bin_boundaries": [10, 11, 12],
                },
                {
                    "length": 5,
                    "unit": "m",
                    "label": "x",
                    "bin_boundaries": [0, 1, 2, 3, 4, 5],
                },
            ],
            "last_metadata_timestamp": 123456,
            "data": [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]],
            "errors": [[5, 4, 3, 2, 1], [10, 9, 8, 7, 6]],
            "info": "info_string",
        }
        buf = serialise_hs01(original_hist)

        hist = deserialise_hs01(buf)
        assert hist["source"] == original_hist["source"]
        assert hist["timestamp"] == original_hist["timestamp"]
        assert hist["current_shape"] == original_hist["current_shape"]
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][0], original_hist["dim_metadata"][0]
        )
        self._check_metadata_for_one_dimension(
            hist["dim_metadata"][1], original_hist["dim_metadata"][1]
        )
        assert np.array_equal(hist["data"], original_hist["data"])
        assert np.array_equal(hist["errors"], original_hist["errors"])
        assert hist["info"] == original_hist["info"]
        assert (
            hist["last_metadata_timestamp"] == original_hist["last_metadata_timestamp"]
        )

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "hs01" in SERIALISERS
        assert "hs01" in DESERIALISERS



from collections import namedtuple
from typing import List, Union

import flatbuffers
from flatbuffers.packer import struct as flatbuffer_struct

from streaming_data_types.fbschemas.forwarder_config_update_rf5k import (
    ConfigUpdate,
    Protocol,
    Stream,
    UpdateType,
)
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"rf5k"

ConfigurationUpdate = namedtuple("ConfigurationUpdate", ("config_change", "streams"))

StreamInfo = namedtuple("StreamInfo", ("channel", "schema", "topic", "protocol"))


def deserialise_rf5k(buffer: Union[bytearray, bytes]) -> ConfigurationUpdate:
    """
    Deserialise FlatBuffer rf5k.

    :param buffer: The FlatBuffers buffer.
    :return: The deserialised data.
    """
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    config_message = ConfigUpdate.ConfigUpdate.GetRootAsConfigUpdate(buffer, 0)

    streams = []
    try:
        for i in range(config_message.StreamsLength()):
            stream_message = config_message.Streams(i)
            streams.append(
                StreamInfo(
                    stream_message.Channel().decode("utf-8")
                    if stream_message.Channel()
                    else "",
                    stream_message.Schema().decode("utf-8")
                    if stream_message.Schema()
                    else "",
                    stream_message.Topic().decode("utf-8")
                    if stream_message.Topic()
                    else "",
                    stream_message.Protocol(),
                )
            )
    except flatbuffer_struct.error:
        pass  # No streams in buffer

    return ConfigurationUpdate(config_message.ConfigChange(), streams)


def serialise_stream(
    builder: flatbuffers.Builder,
    protocol: Protocol,
    channel_offset: int,
    schema_offset: int,
    topic_offset: int,
) -> int:
    Stream.StreamStart(builder)
    Stream.StreamAddProtocol(builder, protocol)
    Stream.StreamAddTopic(builder, topic_offset)
    Stream.StreamAddSchema(builder, schema_offset)
    Stream.StreamAddChannel(builder, channel_offset)
    return Stream.StreamEnd(builder)


def serialise_rf5k(config_change: UpdateType, streams: List[StreamInfo]) -> bytes:
    """
    Serialise config update message as an rf5k FlatBuffers message.

    :param config_change:
    :param streams: channel, schema and output topic configurations
    :return:
    """
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    if streams:
        # We have to use multiple loops/list comprehensions here because we cannot create strings after we have
        # called StreamStart and cannot create streams after we have called StartVector
        stream_field_offsets = [
            (
                builder.CreateString(stream.channel),
                builder.CreateString(stream.schema),
                builder.CreateString(stream.topic),
            )
            for stream in streams
        ]
        stream_offsets = [
            serialise_stream(builder, stream.protocol, *stream_fields)
            for stream, stream_fields in zip(streams, stream_field_offsets)
        ]

        ConfigUpdate.ConfigUpdateStartStreamsVector(builder, len(streams))
        for stream_offset in stream_offsets:
            builder.PrependUOffsetTRelative(stream_offset)
        streams_offset = builder.EndVector(len(streams))

    # Build the actual buffer
    ConfigUpdate.ConfigUpdateStart(builder)
    if streams:
        ConfigUpdate.ConfigUpdateAddStreams(builder, streams_offset)
    ConfigUpdate.ConfigUpdateAddConfigChange(builder, config_change)
    data = ConfigUpdate.ConfigUpdateEnd(builder)

    builder.Finish(data, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class PV_AlarmState(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PV_AlarmState()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPV_AlarmState(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def PV_AlarmStateBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x70\x76\x41\x6C", size_prefixed=size_prefixed
        )

    # PV_AlarmState
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PV_AlarmState
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # PV_AlarmState
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # PV_AlarmState
    def State(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # PV_AlarmState
    def CaState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # PV_AlarmState
    def Severity(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # PV_AlarmState
    def Message(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def PV_AlarmStateStart(builder):
    builder.StartObject(6)


def Start(builder):
    return PV_AlarmStateStart(builder)


def PV_AlarmStateAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def AddSourceName(builder, sourceName):
    return PV_AlarmStateAddSourceName(builder, sourceName)


def PV_AlarmStateAddTimestamp(builder, timestamp):
    builder.PrependInt64Slot(1, timestamp, 0)


def AddTimestamp(builder, timestamp):
    return PV_AlarmStateAddTimestamp(builder, timestamp)


def PV_AlarmStateAddState(builder, state):
    builder.PrependInt16Slot(2, state, 0)


def AddState(builder, state):
    return PV_AlarmStateAddState(builder, state)


def PV_AlarmStateAddCaState(builder, caState):
    builder.PrependInt16Slot(3, caState, 0)


def AddCaState(builder, caState):
    return PV_AlarmStateAddCaState(builder, caState)


def PV_AlarmStateAddSeverity(builder, severity):
    builder.PrependInt16Slot(4, severity, 0)


def AddSeverity(builder, severity):
    return PV_AlarmStateAddSeverity(builder, severity)


def PV_AlarmStateAddMessage(builder, message):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(message), 0
    )


def AddMessage(builder, message):
    return PV_AlarmStateAddMessage(builder, message)


def PV_AlarmStateEnd(builder):
    return builder.EndObject()


def End(builder):
    return PV_AlarmStateEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class EpicsPVConnectionInfo(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = EpicsPVConnectionInfo()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsEpicsPVConnectionInfo(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def EpicsPVConnectionInfoBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x70\x76\x43\x6E", size_prefixed=size_prefixed
        )

    # EpicsPVConnectionInfo
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # EpicsPVConnectionInfo
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int64Flags, o + self._tab.Pos)
        return 0

    # EpicsPVConnectionInfo
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0

    # EpicsPVConnectionInfo
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # EpicsPVConnectionInfo
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def EpicsPVConnectionInfoStart(builder):
    builder.StartObject(4)


def Start(builder):
    return EpicsPVConnectionInfoStart(builder)


def EpicsPVConnectionInfoAddTimestamp(builder, timestamp):
    builder.PrependInt64Slot(0, timestamp, 0)


def AddTimestamp(builder, timestamp):
    return EpicsPVConnectionInfoAddTimestamp(builder, timestamp)


def EpicsPVConnectionInfoAddStatus(builder, status):
    builder.PrependInt16Slot(1, status, 0)


def AddStatus(builder, status):
    return EpicsPVConnectionInfoAddStatus(builder, status)


def EpicsPVConnectionInfoAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def AddSourceName(builder, sourceName):
    return EpicsPVConnectionInfoAddSourceName(builder, sourceName)


def EpicsPVConnectionInfoAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def AddServiceId(builder, serviceId):
    return EpicsPVConnectionInfoAddServiceId(builder, serviceId)


def EpicsPVConnectionInfoEnd(builder):
    return builder.EndObject()


def End(builder):
    return EpicsPVConnectionInfoEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class Float(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsFloat(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Float()
        x.Init(buf, n + offset)
        return x

    # Float
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Float
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0


def FloatStart(builder):
    builder.StartObject(1)


def FloatAddValue(builder, value):
    builder.PrependFloat32Slot(0, value, 0.0)


def FloatEnd(builder):
    return builder.EndObject()



from datetime import datetime, timezone

import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.action_response_answ import (
    ActionOutcome,
    ActionType,
    deserialise_answ,
    serialise_answ,
)
from streaming_data_types.exceptions import WrongSchemaException


class TestSerialisationAnsw:
    def test_serialise_and_deserialise_answ_message(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "service_id": "some_service_id_1234",
            "job_id": "some_job_id_abcdef",
            "command_id": "some command id",
            "action": ActionType.SetStopTime,
            "outcome": ActionOutcome.Failure,
            "message": "some random error message",
            "status_code": 123456789,
            "stop_time": datetime(
                year=2021,
                month=2,
                day=12,
                hour=2,
                minute=12,
                second=12,
                tzinfo=timezone.utc,
            ),
        }

        buf = serialise_answ(**original_entry)
        entry = deserialise_answ(buf)

        assert entry.service_id == original_entry["service_id"]
        assert entry.command_id == original_entry["command_id"]
        assert entry.job_id == original_entry["job_id"]
        assert entry.message == original_entry["message"]
        assert entry.action == original_entry["action"]
        assert entry.outcome == original_entry["outcome"]
        assert entry.status_code == original_entry["status_code"]
        assert entry.stop_time == original_entry["stop_time"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {
            "service_id": "some_service_id_1234",
            "job_id": "some_job_id_abcdef",
            "command_id": "some command id",
            "action": ActionType.SetStopTime,
            "outcome": ActionOutcome.Failure,
            "message": "some random error message",
            "status_code": 123456789,
            "stop_time": datetime(
                year=2021,
                month=2,
                day=12,
                hour=2,
                minute=12,
                second=12,
                tzinfo=timezone.utc,
            ),
        }

        buf = serialise_answ(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_answ(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "answ" in SERIALISERS
        assert "answ" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class RunStart(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsRunStart(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RunStart()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def RunStartBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x70\x6C\x37\x32", size_prefixed=size_prefixed
        )

    # RunStart
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RunStart
    def StartTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # RunStart
    def StopTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # RunStart
    def RunName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def InstrumentName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def NexusStructure(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def JobId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def Broker(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def Filename(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def NPeriods(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 1

    # RunStart
    def DetectorSpectrumMap(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from .SpectraDetectorMapping import SpectraDetectorMapping

            obj = SpectraDetectorMapping()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # RunStart
    def Metadata(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStart
    def ControlTopic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def RunStartStart(builder):
    builder.StartObject(13)


def RunStartAddStartTime(builder, startTime):
    builder.PrependUint64Slot(0, startTime, 0)


def RunStartAddStopTime(builder, stopTime):
    builder.PrependUint64Slot(1, stopTime, 0)


def RunStartAddRunName(builder, runName):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(runName), 0
    )


def RunStartAddInstrumentName(builder, instrumentName):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(instrumentName), 0
    )


def RunStartAddNexusStructure(builder, nexusStructure):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(nexusStructure), 0
    )


def RunStartAddJobId(builder, jobId):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(jobId), 0
    )


def RunStartAddBroker(builder, broker):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(broker), 0
    )


def RunStartAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        7, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def RunStartAddFilename(builder, filename):
    builder.PrependUOffsetTRelativeSlot(
        8, flatbuffers.number_types.UOffsetTFlags.py_type(filename), 0
    )


def RunStartAddNPeriods(builder, nPeriods):
    builder.PrependUint32Slot(9, nPeriods, 1)


def RunStartAddDetectorSpectrumMap(builder, detectorSpectrumMap):
    builder.PrependUOffsetTRelativeSlot(
        10, flatbuffers.number_types.UOffsetTFlags.py_type(detectorSpectrumMap), 0
    )


def RunStartAddMetadata(builder, metadata):
    builder.PrependUOffsetTRelativeSlot(
        11, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0
    )


def RunStartAddControlTopic(builder, controlTopic):
    builder.PrependUOffsetTRelativeSlot(
        12, flatbuffers.number_types.UOffsetTFlags.py_type(controlTopic), 0
    )


def RunStartEnd(builder):
    return builder.EndObject()



from datetime import datetime, timezone

import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.fbschemas.sample_environment_senv.Location import Location
from streaming_data_types.sample_environment_senv import (
    deserialise_senv,
    serialise_senv,
)

entry_1 = {
    "name": "some_name",
    "timestamp": datetime.now(tz=timezone.utc),
    "channel": 42,
    "message_counter": 123456,
    "sample_ts_delta": 0.005,
    "values": np.arange(100, dtype=np.uint16),
    "value_timestamps": np.arange(50) + 1111,
    "ts_location": Location.End,
}

entry_2 = {
    "name": "some_name_other_name",
    "timestamp": datetime.now(tz=timezone.utc),
    "channel": 11,
    "message_counter": 654321,
    "sample_ts_delta": 1.666,
    "values": np.arange(1000, dtype=np.int64),
    "value_timestamps": None,
    "ts_location": Location.Middle,
}


class TestSerialisationSenv:
    @pytest.mark.parametrize("input_entry", [entry_1, entry_2])
    def test_serialises_and_deserialises_senv(self, input_entry):
        original_entry = input_entry
        buf = serialise_senv(**original_entry)
        deserialised_tuple = deserialise_senv(buf)

        assert original_entry["name"] == deserialised_tuple.name
        assert original_entry["timestamp"] == deserialised_tuple.timestamp
        assert original_entry["channel"] == deserialised_tuple.channel
        assert original_entry["message_counter"] == deserialised_tuple.message_counter
        assert original_entry["sample_ts_delta"] == deserialised_tuple.sample_ts_delta
        assert np.array_equal(original_entry["values"], deserialised_tuple.values)
        assert np.array_equal(
            original_entry["value_timestamps"], deserialised_tuple.value_ts
        )
        assert original_entry["values"].dtype == deserialised_tuple.values.dtype
        assert original_entry["ts_location"] == deserialised_tuple.ts_location

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "senv" in SERIALISERS
        assert "senv" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class RunStop(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsRunStop(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RunStop()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def RunStopBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x36\x73\x34\x74", size_prefixed=size_prefixed
        )

    # RunStop
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # RunStop
    def StopTime(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # RunStop
    def RunName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStop
    def JobId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStop
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # RunStop
    def CommandId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def RunStopStart(builder):
    builder.StartObject(5)


def RunStopAddStopTime(builder, stopTime):
    builder.PrependUint64Slot(0, stopTime, 0)


def RunStopAddRunName(builder, runName):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(runName), 0
    )


def RunStopAddJobId(builder, jobId):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(jobId), 0
    )


def RunStopAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def RunStopAddCommandId(builder, commandId):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(commandId), 0
    )


def RunStopEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class DimensionMetaData(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsDimensionMetaData(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = DimensionMetaData()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def DimensionMetaDataBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x68\x73\x30\x31", size_prefixed=size_prefixed
        )

    # DimensionMetaData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # DimensionMetaData
    def Length(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # DimensionMetaData
    def Unit(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # DimensionMetaData
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # DimensionMetaData
    def BinBoundariesType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # DimensionMetaData
    def BinBoundaries(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None


def DimensionMetaDataStart(builder):
    builder.StartObject(5)


def DimensionMetaDataAddLength(builder, length):
    builder.PrependInt32Slot(0, length, 0)


def DimensionMetaDataAddUnit(builder, unit):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(unit), 0
    )


def DimensionMetaDataAddLabel(builder, label):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0
    )


def DimensionMetaDataAddBinBoundariesType(builder, binBoundariesType):
    builder.PrependUint8Slot(3, binBoundariesType, 0)


def DimensionMetaDataAddBinBoundaries(builder, binBoundaries):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(binBoundaries), 0
    )


def DimensionMetaDataEnd(builder):
    return builder.EndObject()



import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.status_x5f2 import deserialise_x5f2, serialise_x5f2

original_entry = {
    "software_name": "nicos/test",
    "software_version": "1.0.0",
    "service_id": "1a2b3c",
    "host_name": "localhost",
    "process_id": 1234,
    "update_interval": 0,
    "status_json": '{"content" : "log_or_status_message"}',
}


class TestSerialisationX52f:
    def test_serialises_and_deserialises_x5f2_message_correctly(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        buf = serialise_x5f2(**original_entry)
        entry = deserialise_x5f2(buf)

        assert entry.software_name == original_entry["software_name"]
        assert entry.software_version == original_entry["software_version"]
        assert entry.service_id == original_entry["service_id"]
        assert entry.host_name == original_entry["host_name"]
        assert entry.process_id == original_entry["process_id"]
        assert entry.update_interval == original_entry["update_interval"]
        assert entry.status_json == original_entry["status_json"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_x5f2(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_x5f2(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "x5f2" in SERIALISERS
        assert "x5f2" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FB_Tables

import flatbuffers


class epicsTimeStamp(object):
    __slots__ = ["_tab"]

    # epicsTimeStamp
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # epicsTimeStamp
    def SecPastEpoch(self):
        return self._tab.Get(
            flatbuffers.number_types.Int32Flags,
            self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0),
        )

    # epicsTimeStamp
    def Nsec(self):
        return self._tab.Get(
            flatbuffers.number_types.Int32Flags,
            self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4),
        )


def CreateepicsTimeStamp(builder, secPastEpoch, nsec):
    builder.Prep(4, 8)
    builder.PrependInt32(nsec)
    builder.PrependInt32(secPastEpoch)
    return builder.Offset()



import pytest
from datetime import datetime, timezone
from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types import (
    deserialise_scal,
    serialise_scal,
)
import numpy as np
from streaming_data_types.exceptions import WrongSchemaException


class TestSerialisation_scal:
    original_entry = {
        "source_name": "test_source",
        "timestamp": datetime(
            year=2021,
            month=3,
            day=26,
            hour=12,
            minute=32,
            second=11,
            tzinfo=timezone.utc,
        ),
    }

    @pytest.mark.parametrize("value", [1234, 3.14])
    def test_serialises_and_deserialises_scal_message(self, value):
        buf = serialise_scal(**self.original_entry, value=value)
        deserialised_tuple = deserialise_scal(buf)

        assert deserialised_tuple.source_name == self.original_entry["source_name"]
        assert deserialised_tuple.timestamp == self.original_entry["timestamp"]
        assert deserialised_tuple.value == value

    @pytest.mark.parametrize(
        "value_type",
        [
            np.dtype("int8"),
            np.dtype("uint8"),
            np.dtype("int16"),
            np.dtype("uint16"),
            np.dtype("int32"),
            np.dtype("uint32"),
            np.dtype("int64"),
            np.dtype("uint64"),
            np.dtype("float32"),
            np.dtype("float64"),
        ],
    )
    def test_serialises_and_deserialises_scal_message_numpy(self, value_type):
        value = np.arange(10, dtype=value_type)
        buf = serialise_scal(**self.original_entry, value=value)
        deserialised_tuple = deserialise_scal(buf)

        assert deserialised_tuple.source_name == self.original_entry["source_name"]
        assert deserialised_tuple.timestamp == self.original_entry["timestamp"]
        assert (deserialised_tuple.value == value).all()
        assert deserialised_tuple.value.dtype == value_type

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_scal(**self.original_entry, value=123)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_scal(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "scal" in SERIALISERS
        assert "scal" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

from __future__ import absolute_import, division, print_function

import flatbuffers


# /// pylint: skip-file
class CacheEntry(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsCacheEntry(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = CacheEntry()
        x.Init(buf, n + offset)
        return x

    # CacheEntry
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # CacheEntry
    def Key(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return ""

    # CacheEntry
    def Time(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float64Flags, o + self._tab.Pos
            )
        return 0.0

    # CacheEntry
    def Ttl(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float64Flags, o + self._tab.Pos
            )
        return 0.0

    # CacheEntry
    def Expired(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
        return 0

    # CacheEntry
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return ""


def CacheEntryStart(builder):
    builder.StartObject(5)


def CacheEntryAddKey(builder, key):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(key), 0
    )


def CacheEntryAddTime(builder, time):
    builder.PrependFloat64Slot(1, time, 0.0)


def CacheEntryAddTtl(builder, ttl):
    builder.PrependFloat64Slot(2, ttl, 0.0)


def CacheEntryAddExpired(builder, expired):
    builder.PrependBoolSlot(3, expired, 0)


def CacheEntryAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def CacheEntryEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Attribute(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsAttribute(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Attribute()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def AttributeBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x41\x44\x41\x72", size_prefixed=size_prefixed
        )

    # Attribute
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Attribute
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Attribute
    def Description(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Attribute
    def Source(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Attribute
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Attribute
    def Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Attribute
    def DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Attribute
    def DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Attribute
    def DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0


def AttributeStart(builder):
    builder.StartObject(5)


def AttributeAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0
    )


def AttributeAddDescription(builder, description):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0
    )


def AttributeAddSource(builder, source):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(source), 0
    )


def AttributeAddDataType(builder, dataType):
    builder.PrependInt8Slot(3, dataType, 0)


def AttributeAddData(builder, data):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0
    )


def AttributeStartDataVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def AttributeEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class Short(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsShort(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Short()
        x.Init(buf, n + offset)
        return x

    # Short
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Short
    def Value(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int16Flags, o + self._tab.Pos)
        return 0


def ShortStart(builder):
    builder.StartObject(1)


def ShortAddValue(builder, value):
    builder.PrependInt16Slot(0, value, 0)


def ShortEnd(builder):
    return builder.EndObject()



import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.nicos_cache_ns10 import deserialise_ns10, serialise_ns10


class TestSerialisationNs10:
    def test_serialises_and_deserialises_ns10_message_correctly(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "key": "some_key",
            "time_stamp": 123456,
            "ttl": 567890,
            "expired": True,
            "value": "some_value",
        }

        buf = serialise_ns10(**original_entry)
        entry = deserialise_ns10(buf)

        assert entry.key == original_entry["key"]
        assert entry.time_stamp == original_entry["time_stamp"]
        assert entry.ttl == original_entry["ttl"]
        assert entry.expired == original_entry["expired"]
        assert entry.value == original_entry["value"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {
            "key": "some_key",
            "time_stamp": 123456,
            "ttl": 567890,
            "expired": True,
            "value": "some_value",
        }
        buf = serialise_ns10(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_ns10(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "ns10" in SERIALISERS
        assert "ns10" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class FinishedWriting(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsFinishedWriting(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = FinishedWriting()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def FinishedWritingBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x77\x72\x64\x6E", size_prefixed=size_prefixed
        )

    # FinishedWriting
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # FinishedWriting
    def ServiceId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # FinishedWriting
    def JobId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # FinishedWriting
    def ErrorEncountered(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # FinishedWriting
    def FileName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # FinishedWriting
    def Metadata(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # FinishedWriting
    def Message(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def FinishedWritingStart(builder):
    builder.StartObject(6)


def FinishedWritingAddServiceId(builder, serviceId):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(serviceId), 0
    )


def FinishedWritingAddJobId(builder, jobId):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(jobId), 0
    )


def FinishedWritingAddErrorEncountered(builder, errorEncountered):
    builder.PrependBoolSlot(2, errorEncountered, 0)


def FinishedWritingAddFileName(builder, fileName):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(fileName), 0
    )


def FinishedWritingAddMetadata(builder, metadata):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(metadata), 0
    )


def FinishedWritingAddMessage(builder, message):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(message), 0
    )


def FinishedWritingEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FB_Tables


class DType(object):
    Int8 = 0
    Uint8 = 1
    Int16 = 2
    Uint16 = 3
    Int32 = 4
    Uint32 = 5
    Int64 = 6
    Uint64 = 7
    Float32 = 8
    Float64 = 9
    c_string = 10



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class Value(object):
    NONE = 0
    Byte = 1
    UByte = 2
    Short = 3
    UShort = 4
    Int = 5
    UInt = 6
    Long = 7
    ULong = 8
    Float = 9
    Double = 10
    ArrayByte = 11
    ArrayUByte = 12
    ArrayShort = 13
    ArrayUShort = 14
    ArrayInt = 15
    ArrayUInt = 16
    ArrayLong = 17
    ArrayULong = 18
    ArrayFloat = 19
    ArrayDouble = 20
    String = 21
    ArrayString = 22



from collections import namedtuple

import flatbuffers
import numpy as np

import streaming_data_types.fbschemas.eventdata_ev43.Event43Message as Event43Message
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"ev43"


EventData = namedtuple(
    "EventData",
    (
        "source_name",
        "message_id",
        "pulse_time",
        "pulse_index",
        "time_of_flight",
        "detector_id",
    ),
)


def deserialise_ev43(buffer):
    """
    Deserialise FlatBuffer ev43.

    :param buffer: The FlatBuffers buffer.
    :return: The deserialised data.
    """
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    event = Event43Message.Event43Message.GetRootAsEvent43Message(buffer, 0)

    return EventData(
        event.SourceName().decode("utf-8"),
        event.MessageId(),
        event.PulseTimeAsNumpy(),
        event.PulseIndexAsNumpy(),
        event.TimeOfFlightAsNumpy(),
        event.DetectorIdAsNumpy(),
    )


def serialise_ev43(
    source_name, message_id, pulse_time, pulse_index, time_of_flight, detector_id
):
    """
    Serialise event data as an ev43 FlatBuffers message.

    :param source_name:
    :param message_id:
    :param pulse_time:
    :param pulse_index:
    :param time_of_flight:
    :param detector_id:
    :return:
    """
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    source = builder.CreateString(source_name)

    pulse_ts_data = builder.CreateNumpyVector(np.asarray(pulse_time).astype(np.uint64))
    pulse_ix_data = builder.CreateNumpyVector(np.asarray(pulse_index).astype(np.uint32))
    tof_data = builder.CreateNumpyVector(np.asarray(time_of_flight).astype(np.uint32))
    det_data = builder.CreateNumpyVector(np.asarray(detector_id).astype(np.uint32))

    # Build the actual buffer
    Event43Message.Event43MessageStart(builder)
    Event43Message.Event43MessageAddPulseTime(builder, pulse_ts_data)
    Event43Message.Event43MessageAddPulseIndex(builder, pulse_ix_data)
    Event43Message.Event43MessageAddDetectorId(builder, det_data)
    Event43Message.Event43MessageAddTimeOfFlight(builder, tof_data)
    Event43Message.Event43MessageAddMessageId(builder, message_id)
    Event43Message.Event43MessageAddSourceName(builder, source)

    data = Event43Message.Event43MessageEnd(builder)

    builder.Finish(data, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())



import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.area_detector_NDAr import deserialise_ndar, serialise_ndar
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.fbschemas.NDAr_NDArray_schema.DType import DType


class TestSerialisationNDAr:
    def test_serialises_and_deserialises_NDAr_message_correctly_float64_1_pixel(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "id": 754,
            "dims": [1, 1],
            "data_type": DType.Float64,
            "data": [54, 78, 100, 156, 43, 1, 23, 0],
        }

        buf = serialise_ndar(**original_entry)
        entry = deserialise_ndar(buf)

        assert entry.id == original_entry["id"]
        assert np.array_equal(entry.data, [[3.1991794446845865e-308]])

    def test_serialises_and_deserialises_NDAr_message_correctly_int32_3_pixel(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "id": 754,
            "dims": [1, 3],
            "data_type": DType.Int32,
            "data": [54, 78, 100, 200, 32, 19, 2, 156, 43, 1, 23, 0],
        }

        buf = serialise_ndar(**original_entry)
        entry = deserialise_ndar(buf)

        assert entry.id == original_entry["id"]
        assert np.array_equal(entry.data, [[-932950474, -1677585632, 1507627]])

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {
            "id": 754,
            "dims": [10, 10],
            "data_type": 0,
            "data": [0, 0, 100, 200, 250],
        }

        buf = serialise_ndar(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_ndar(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "NDAr" in SERIALISERS
        assert "NDAr" in DESERIALISERS



import pytest

from streaming_data_types.exceptions import ShortBufferException
from streaming_data_types.utils import check_schema_identifier


def test_schema_check_throws_if_buffer_too_short():
    short_buffer = b"1234567"
    with pytest.raises(ShortBufferException):
        check_schema_identifier(short_buffer, b"1234")



from streaming_data_types._version import version
from streaming_data_types.action_response_answ import deserialise_answ, serialise_answ
from streaming_data_types.area_detector_ADAr import deserialise_ADAr, serialise_ADAr
from streaming_data_types.area_detector_NDAr import deserialise_ndar, serialise_ndar
from streaming_data_types.epics_connection_info_ep00 import (
    deserialise_ep00,
    serialise_ep00,
)
from streaming_data_types.eventdata_ev42 import deserialise_ev42, serialise_ev42
from streaming_data_types.eventdata_ev43 import deserialise_ev43, serialise_ev43
from streaming_data_types.finished_writing_wrdn import deserialise_wrdn, serialise_wrdn
from streaming_data_types.forwarder_config_update_rf5k import (
    deserialise_rf5k,
    serialise_rf5k,
)
from streaming_data_types.histogram_hs00 import deserialise_hs00, serialise_hs00
from streaming_data_types.histogram_hs01 import deserialise_hs01, serialise_hs01
from streaming_data_types.logdata_f142 import deserialise_f142, serialise_f142
from streaming_data_types.nicos_cache_ns10 import deserialise_ns10, serialise_ns10
from streaming_data_types.run_start_pl72 import deserialise_pl72, serialise_pl72
from streaming_data_types.run_stop_6s4t import deserialise_6s4t, serialise_6s4t
from streaming_data_types.sample_environment_senv import (
    deserialise_senv,
    serialise_senv,
)
from streaming_data_types.status_x5f2 import deserialise_x5f2, serialise_x5f2
from streaming_data_types.timestamps_tdct import deserialise_tdct, serialise_tdct
from streaming_data_types.epics_pv_conn_status_pvCn import (
    deserialise_pvCn,
    serialise_pvCn,
)
from streaming_data_types.epics_pv_alarm_status_pvAl import (
    deserialise_pvAl,
    serialise_pvAl,
)
from streaming_data_types.epics_pv_scalar_data_scal import (
    deserialise_scal,
    serialise_scal,
)

__version__ = version

SERIALISERS = {
    "ev42": serialise_ev42,
    "ev43": serialise_ev43,
    "hs00": serialise_hs00,
    "hs01": serialise_hs01,
    "f142": serialise_f142,
    "ns10": serialise_ns10,
    "pl72": serialise_pl72,
    "6s4t": serialise_6s4t,
    "x5f2": serialise_x5f2,
    "ep00": serialise_ep00,
    "tdct": serialise_tdct,
    "rf5k": serialise_rf5k,
    "answ": serialise_answ,
    "wrdn": serialise_wrdn,
    "NDAr": serialise_ndar,
    "senv": serialise_senv,
    "ADAr": serialise_ADAr,
    "pvCn": serialise_pvCn,
    "pvAl": serialise_pvAl,
    "scal": serialise_scal,
}


DESERIALISERS = {
    "ev42": deserialise_ev42,
    "ev43": deserialise_ev43,
    "hs00": deserialise_hs00,
    "hs01": deserialise_hs01,
    "f142": deserialise_f142,
    "ns10": deserialise_ns10,
    "pl72": deserialise_pl72,
    "6s4t": deserialise_6s4t,
    "x5f2": deserialise_x5f2,
    "ep00": deserialise_ep00,
    "tdct": deserialise_tdct,
    "rf5k": deserialise_rf5k,
    "answ": deserialise_answ,
    "wrdn": deserialise_wrdn,
    "NDAr": deserialise_ndar,
    "senv": deserialise_senv,
    "ADAr": deserialise_ADAr,
    "pvCn": deserialise_pvCn,
    "pvAl": deserialise_pvAl,
    "scal": deserialise_scal,
}



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class Stream(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsStream(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Stream()
        x.Init(buf, n + offset)
        return x

    # Stream
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Stream
    def Channel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Stream
    def Schema(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Stream
    def Topic(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Stream
    def Protocol(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0


def StreamStart(builder):
    builder.StartObject(4)


def StreamAddChannel(builder, channel):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(channel), 0
    )


def StreamAddSchema(builder, schema):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(schema), 0
    )


def StreamAddTopic(builder, topic):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(topic), 0
    )


def StreamAddProtocol(builder, protocol):
    builder.PrependUint16Slot(3, protocol, 0)


def StreamEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FB_Tables

import flatbuffers


class NDArray(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsNDArray(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = NDArray()
        x.Init(buf, n + offset)
        return x

    # NDArray
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # NDArray
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # NDArray
    def TimeStamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float64Flags, o + self._tab.Pos
            )
        return 0.0

    # NDArray
    def EpicsTS(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .epicsTimeStamp import epicsTimeStamp

            obj = epicsTimeStamp()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # NDArray
    def Dims(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # NDArray
    def DimsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # NDArray
    def DimsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NDArray
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # NDArray
    def PData(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # NDArray
    def PDataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # NDArray
    def PDataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NDArray
    def PAttributeList(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .NDAttribute import NDAttribute

            obj = NDAttribute()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # NDArray
    def PAttributeListLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0


def NDArrayStart(builder):
    builder.StartObject(7)


def NDArrayAddId(builder, id):
    builder.PrependInt32Slot(0, id, 0)


def NDArrayAddTimeStamp(builder, timeStamp):
    builder.PrependFloat64Slot(1, timeStamp, 0.0)


def NDArrayAddEpicsTS(builder, epicsTS):
    builder.PrependStructSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(epicsTS), 0
    )


def NDArrayAddDims(builder, dims):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(dims), 0
    )


def NDArrayStartDimsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def NDArrayAddDataType(builder, dataType):
    builder.PrependInt8Slot(4, dataType, 0)


def NDArrayAddPData(builder, pData):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(pData), 0
    )


def NDArrayStartPDataVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def NDArrayAddPAttributeList(builder, pAttributeList):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(pAttributeList), 0
    )


def NDArrayStartPAttributeListVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def NDArrayEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class ISISData(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsISISData(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ISISData()
        x.Init(buf, n + offset)
        return x

    # ISISData
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ISISData
    def PeriodNumber(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # ISISData
    def RunState(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # ISISData
    def ProtonCharge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 0.0


def ISISDataStart(builder):
    builder.StartObject(3)


def ISISDataAddPeriodNumber(builder, periodNumber):
    builder.PrependUint32Slot(0, periodNumber, 0)


def ISISDataAddRunState(builder, runState):
    builder.PrependInt8Slot(1, runState, 0)


def ISISDataAddProtonCharge(builder, protonCharge):
    builder.PrependFloat32Slot(2, protonCharge, 0.0)


def ISISDataEnd(builder):
    return builder.EndObject()



import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.eventdata_ev43 import deserialise_ev43, serialise_ev43
from streaming_data_types.exceptions import WrongSchemaException


class TestSerialisationEv42:
    def test_serialises_and_deserialises_ev43_message_correctly(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "source_name": "some_source",
            "message_id": 123456,
            "pulse_time": [567890, 568890],
            "pulse_index": [0, 4],
            "time_of_flight": [1, 2, 3, 4, 5, 6, 7, 8, 9],
            "detector_id": [10, 20, 30, 40, 50, 60, 70, 80, 90],
        }

        buf = serialise_ev43(**original_entry)
        entry = deserialise_ev43(buf)

        assert entry.source_name == original_entry["source_name"]
        assert entry.message_id == original_entry["message_id"]
        assert np.array_equal(entry.pulse_time, original_entry["pulse_time"])
        assert np.array_equal(entry.pulse_index, original_entry["pulse_index"])
        assert np.array_equal(entry.time_of_flight, original_entry["time_of_flight"])
        assert np.array_equal(entry.detector_id, original_entry["detector_id"])

    def test_serialises_and_deserialises_ev43_message_correctly_for_numpy_arrays(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "source_name": "some_source",
            "message_id": 123456,
            "pulse_time": np.array([567890, 568890]),
            "pulse_index": np.array([0, 4]),
            "time_of_flight": np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]),
            "detector_id": np.array([10, 20, 30, 40, 50, 60, 70, 80, 90]),
        }

        buf = serialise_ev43(**original_entry)
        entry = deserialise_ev43(buf)

        assert entry.source_name == original_entry["source_name"]
        assert entry.message_id == original_entry["message_id"]
        assert np.array_equal(entry.pulse_time, original_entry["pulse_time"])
        assert np.array_equal(entry.pulse_index, original_entry["pulse_index"])
        assert np.array_equal(entry.time_of_flight, original_entry["time_of_flight"])
        assert np.array_equal(entry.detector_id, original_entry["detector_id"])

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {
            "source_name": "some_source",
            "message_id": 123456,
            "pulse_time": [567890, 568890],
            "pulse_index": [0, 4],
            "time_of_flight": [1, 2, 3, 4, 5, 6, 7, 8, 9],
            "detector_id": [10, 20, 30, 40, 50, 60, 70, 80, 90],
        }
        buf = serialise_ev43(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_ev43(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "ev43" in SERIALISERS
        assert "ev43" in DESERIALISERS



import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.fbschemas.forwarder_config_update_rf5k.UpdateType import (
    UpdateType,
)
from streaming_data_types.forwarder_config_update_rf5k import (
    Protocol,
    StreamInfo,
    deserialise_rf5k,
    serialise_rf5k,
)


class TestSerialisationRf5k:
    def test_serialises_and_deserialises_rf5k_message_with_streams_correctly(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        stream_1 = StreamInfo("channel1", "f142", "topic1", Protocol.Protocol.PVA)
        stream_2 = StreamInfo("channel2", "TdcTime", "topic2", Protocol.Protocol.CA)
        original_entry = {
            "config_change": UpdateType.ADD,
            "streams": [stream_1, stream_2],
        }

        buf = serialise_rf5k(**original_entry)
        entry = deserialise_rf5k(buf)

        assert entry.config_change == original_entry["config_change"]
        assert stream_1 in entry.streams
        assert stream_2 in entry.streams

    def test_serialises_and_deserialises_rf5k_message_without_streams_correctly(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {"config_change": UpdateType.REMOVEALL, "streams": []}

        buf = serialise_rf5k(**original_entry)
        entry = deserialise_rf5k(buf)

        assert entry.config_change == original_entry["config_change"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {"config_change": UpdateType.REMOVEALL, "streams": []}

        buf = serialise_rf5k(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_rf5k(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "rf5k" in SERIALISERS
        assert "rf5k" in DESERIALISERS



import flatbuffers
import numpy

import streaming_data_types.fbschemas.histogram_hs00.ArrayDouble as ArrayDouble
import streaming_data_types.fbschemas.histogram_hs00.ArrayFloat as ArrayFloat
import streaming_data_types.fbschemas.histogram_hs00.ArrayUInt as ArrayUInt
import streaming_data_types.fbschemas.histogram_hs00.ArrayULong as ArrayULong
import streaming_data_types.fbschemas.histogram_hs00.DimensionMetaData as DimensionMetaData
import streaming_data_types.fbschemas.histogram_hs00.EventHistogram as EventHistogram
from streaming_data_types.fbschemas.histogram_hs00.Array import Array
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"hs00"


_array_for_type = {
    Array.ArrayUInt: ArrayUInt.ArrayUInt(),
    Array.ArrayULong: ArrayULong.ArrayULong(),
    Array.ArrayFloat: ArrayFloat.ArrayFloat(),
}


def _create_array_object_for_type(array_type):
    return _array_for_type.get(array_type, ArrayDouble.ArrayDouble())


def deserialise_hs00(buffer):
    """
    Deserialise flatbuffer hs10 into a histogram.

    :param buffer:
    :return: dict of histogram information
    """
    check_schema_identifier(buffer, FILE_IDENTIFIER)
    event_hist = EventHistogram.EventHistogram.GetRootAsEventHistogram(buffer, 0)

    dims = []
    for i in range(event_hist.DimMetadataLength()):
        bins_fb = _create_array_object_for_type(
            event_hist.DimMetadata(i).BinBoundariesType()
        )

        # Get bins
        bins_offset = event_hist.DimMetadata(i).BinBoundaries()
        bins_fb.Init(bins_offset.Bytes, bins_offset.Pos)
        bin_boundaries = bins_fb.ValueAsNumpy()

        hist_info = {
            "length": event_hist.DimMetadata(i).Length(),
            "bin_boundaries": bin_boundaries,
            "unit": event_hist.DimMetadata(i).Unit().decode("utf-8")
            if event_hist.DimMetadata(i).Unit()
            else "",
            "label": event_hist.DimMetadata(i).Label().decode("utf-8")
            if event_hist.DimMetadata(i).Label()
            else "",
        }
        dims.append(hist_info)

    metadata_timestamp = event_hist.LastMetadataTimestamp()

    data_fb = _create_array_object_for_type(event_hist.DataType())
    data_offset = event_hist.Data()
    data_fb.Init(data_offset.Bytes, data_offset.Pos)
    shape = event_hist.CurrentShapeAsNumpy().tolist()
    data = data_fb.ValueAsNumpy().reshape(shape)

    # Get the errors
    errors_offset = event_hist.Errors()
    if errors_offset:
        errors_fb = _create_array_object_for_type(event_hist.ErrorsType())
        errors_fb.Init(errors_offset.Bytes, errors_offset.Pos)
        errors = errors_fb.ValueAsNumpy().reshape(shape)
    else:
        errors = []

    hist = {
        "source": event_hist.Source().decode("utf-8") if event_hist.Source() else "",
        "timestamp": event_hist.Timestamp(),
        "current_shape": shape,
        "dim_metadata": dims,
        "data": data,
        "errors": errors,
        "last_metadata_timestamp": metadata_timestamp,
        "info": event_hist.Info().decode("utf-8") if event_hist.Info() else "",
    }
    return hist


def _serialise_metadata(builder, length, edges, unit, label):
    unit_offset = builder.CreateString(unit)
    label_offset = builder.CreateString(label)

    bins_offset, bin_type = _serialise_array(builder, edges)

    DimensionMetaData.DimensionMetaDataStart(builder)
    DimensionMetaData.DimensionMetaDataAddLength(builder, length)
    DimensionMetaData.DimensionMetaDataAddBinBoundaries(builder, bins_offset)
    DimensionMetaData.DimensionMetaDataAddBinBoundariesType(builder, bin_type)
    DimensionMetaData.DimensionMetaDataAddLabel(builder, label_offset)
    DimensionMetaData.DimensionMetaDataAddUnit(builder, unit_offset)
    return DimensionMetaData.DimensionMetaDataEnd(builder)


def serialise_hs00(histogram):
    """
    Serialise a histogram as an hs00 FlatBuffers message.

    If arrays are provided as numpy arrays with type np.uint32, np.uint64, np.float32
    or np.float64 then type is preserved in output buffer.

    :param histogram: A dictionary containing the histogram to serialise.
    """
    source_offset = None
    info_offset = None

    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)
    if "source" in histogram:
        source_offset = builder.CreateString(histogram["source"])
    if "info" in histogram:
        info_offset = builder.CreateString(histogram["info"])

    # Build shape array
    shape_offset = builder.CreateNumpyVector(
        numpy.array(histogram["current_shape"]).astype(numpy.uint32)
    )

    # Build dimensions metadata
    metadata = []
    for meta in histogram["dim_metadata"]:
        unit = "" if "unit" not in meta else meta["unit"]
        label = "" if "label" not in meta else meta["label"]
        metadata.append(
            _serialise_metadata(
                builder, meta["length"], meta["bin_boundaries"], unit, label
            )
        )

    rank = len(histogram["current_shape"])
    EventHistogram.EventHistogramStartDimMetadataVector(builder, rank)
    # FlatBuffers builds arrays backwards
    for m in reversed(metadata):
        builder.PrependUOffsetTRelative(m)
    metadata_vector = builder.EndVector(rank)

    # Build the data
    data_offset, data_type = _serialise_array(builder, histogram["data"])

    errors_offset = None
    if "errors" in histogram:
        errors_offset, error_type = _serialise_array(builder, histogram["errors"])

    # Build the actual buffer
    EventHistogram.EventHistogramStart(builder)
    if info_offset:
        EventHistogram.EventHistogramAddInfo(builder, info_offset)
    EventHistogram.EventHistogramAddData(builder, data_offset)
    EventHistogram.EventHistogramAddCurrentShape(builder, shape_offset)
    EventHistogram.EventHistogramAddDimMetadata(builder, metadata_vector)
    EventHistogram.EventHistogramAddTimestamp(builder, histogram["timestamp"])
    if source_offset:
        EventHistogram.EventHistogramAddSource(builder, source_offset)
    EventHistogram.EventHistogramAddDataType(builder, data_type)
    if errors_offset:
        EventHistogram.EventHistogramAddErrors(builder, errors_offset)
        EventHistogram.EventHistogramAddErrorsType(builder, error_type)
    if "last_metadata_timestamp" in histogram:
        EventHistogram.EventHistogramAddLastMetadataTimestamp(
            builder, histogram["last_metadata_timestamp"]
        )
    hist_message = EventHistogram.EventHistogramEnd(builder)

    builder.Finish(hist_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


def _serialise_array(builder, data):
    flattened_data = numpy.asarray(data).flatten()

    # Carefully preserve explicitly supported types
    if numpy.issubdtype(flattened_data.dtype, numpy.uint32):
        return _serialise_uint32(builder, flattened_data)
    if numpy.issubdtype(flattened_data.dtype, numpy.uint64):
        return _serialise_uint64(builder, flattened_data)
    if numpy.issubdtype(flattened_data.dtype, numpy.float32):
        return _serialise_float(builder, flattened_data)
    if numpy.issubdtype(flattened_data.dtype, numpy.float64):
        return _serialise_double(builder, flattened_data)

    # Otherwise if it looks like an int then use uint64, or use double as last resort
    if numpy.issubdtype(flattened_data.dtype, numpy.int64):
        return _serialise_uint64(builder, flattened_data)

    return _serialise_double(builder, flattened_data)


def _serialise_float(builder, flattened_data):
    data_type = Array.ArrayFloat
    data_vector = builder.CreateNumpyVector(flattened_data)
    ArrayFloat.ArrayFloatStart(builder)
    ArrayFloat.ArrayFloatAddValue(builder, data_vector)
    data_offset = ArrayFloat.ArrayFloatEnd(builder)
    return data_offset, data_type


def _serialise_double(builder, flattened_data):
    data_type = Array.ArrayDouble
    data_vector = builder.CreateNumpyVector(flattened_data)
    ArrayDouble.ArrayDoubleStart(builder)
    ArrayDouble.ArrayDoubleAddValue(builder, data_vector)
    data_offset = ArrayDouble.ArrayDoubleEnd(builder)
    return data_offset, data_type


def _serialise_uint32(builder, flattened_data):
    data_type = Array.ArrayUInt
    data_vector = builder.CreateNumpyVector(flattened_data)
    ArrayUInt.ArrayUIntStart(builder)
    ArrayUInt.ArrayUIntAddValue(builder, data_vector)
    data_offset = ArrayUInt.ArrayUIntEnd(builder)
    return data_offset, data_type


def _serialise_uint64(builder, flattened_data):
    data_type = Array.ArrayULong
    data_vector = builder.CreateNumpyVector(flattened_data)
    ArrayULong.ArrayULongStart(builder)
    ArrayULong.ArrayULongAddValue(builder, data_vector)
    data_offset = ArrayULong.ArrayULongEnd(builder)
    return data_offset, data_type



import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.timestamps_tdct import deserialise_tdct, serialise_tdct


class TestSerialisationTdct:
    original_entry = {
        "name": "some_name",
        "timestamps": [0, 1, 2, 3, 4],
        "sequence_counter": 42,
    }

    def test_serialises_and_deserialises_tdct_message_with_list_of_timestamps(self):
        buf = serialise_tdct(**self.original_entry)
        deserialised_tuple = deserialise_tdct(buf)

        assert deserialised_tuple.name == self.original_entry["name"]
        assert np.allclose(
            deserialised_tuple.timestamps, np.array(self.original_entry["timestamps"])
        )
        assert (
            deserialised_tuple.sequence_counter
            == self.original_entry["sequence_counter"]
        )

    def test_serialises_and_deserialises_tdct_message_with_array_of_timestamps(self):
        original_entry = {"name": "some_name", "timestamps": np.array([0, 1, 2, 3, 4])}

        buf = serialise_tdct(**original_entry)
        deserialised_tuple = deserialise_tdct(buf)

        assert deserialised_tuple.name == original_entry["name"]
        assert np.allclose(deserialised_tuple.timestamps, original_entry["timestamps"])

    def test_serialises_and_deserialises_tdct_message_with_single_timestamp(self):
        original_entry = {"name": "some_name", "timestamps": np.array(0)}

        buf = serialise_tdct(**original_entry)
        deserialised_tuple = deserialise_tdct(buf)

        assert deserialised_tuple.name == original_entry["name"]
        assert np.allclose(deserialised_tuple.timestamps, original_entry["timestamps"])

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_tdct(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_tdct(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "tdct" in SERIALISERS
        assert "tdct" in DESERIALISERS



import pytest
from datetime import datetime, timezone
from copy import deepcopy
from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types import (
    deserialise_pvCn,
    serialise_pvCn,
)
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.epics_pv_conn_status_pvCn import ConnectionInfo


class TestSerialisation_pvCn:
    original_entry = {
        "timestamp": datetime(
            year=2021,
            month=3,
            day=26,
            hour=12,
            minute=32,
            second=11,
            tzinfo=timezone.utc,
        ),
        "status": ConnectionInfo.DESTROYED,
        "source_name": "test_source",
        "service_id": "test_service",
    }

    def test_serialises_and_deserialises_pvCn_message_1(self):
        buf = serialise_pvCn(**self.original_entry)
        deserialised_tuple = deserialise_pvCn(buf)

        assert deserialised_tuple.timestamp == self.original_entry["timestamp"]
        assert deserialised_tuple.status == self.original_entry["status"]
        assert deserialised_tuple.source_name == self.original_entry["source_name"]
        assert deserialised_tuple.service_id == self.original_entry["service_id"]

    def test_serialises_and_deserialises_pvCn_message_2(self):
        modified_entry = deepcopy(self.original_entry)
        modified_entry.pop("service_id")
        buf = serialise_pvCn(**modified_entry)
        deserialised_tuple = deserialise_pvCn(buf)

        assert deserialised_tuple.timestamp == modified_entry["timestamp"]
        assert deserialised_tuple.status == modified_entry["status"]
        assert deserialised_tuple.source_name == modified_entry["source_name"]
        assert deserialised_tuple.service_id is None

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_pvCn(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_pvCn(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "pvCn" in SERIALISERS
        assert "pvCn" in DESERIALISERS



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class ArrayDouble(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsArrayDouble(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ArrayDouble()
        x.Init(buf, n + offset)
        return x

    # ArrayDouble
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ArrayDouble
    def Value(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Float64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # ArrayDouble
    def ValueAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Float64Flags, o)
        return 0

    # ArrayDouble
    def ValueLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0


def ArrayDoubleStart(builder):
    builder.StartObject(1)


def ArrayDoubleAddValue(builder, value):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(value), 0
    )


def ArrayDoubleStartValueVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def ArrayDoubleEnd(builder):
    return builder.EndObject()



from typing import List, NamedTuple, Optional, Union

import flatbuffers
import numpy as np

from streaming_data_types.fbschemas.timestamps_tdct.timestamp import (
    timestamp,
    timestampAddName,
    timestampAddSequenceCounter,
    timestampAddTimestamps,
    timestampEnd,
    timestampStart,
)
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"tdct"


def serialise_tdct(
    name: str,
    timestamps: Union[np.ndarray, List],
    sequence_counter: Optional[int] = None,
) -> bytes:
    builder = flatbuffers.Builder(1024)
    builder.ForceDefaults(True)

    timestamps = np.atleast_1d(np.asarray(timestamps)).astype(np.uint64)

    name_offset = builder.CreateString(name)

    array_offset = builder.CreateNumpyVector(timestamps)

    timestampStart(builder)
    timestampAddName(builder, name_offset)
    timestampAddTimestamps(builder, array_offset)
    if sequence_counter is not None:
        timestampAddSequenceCounter(builder, sequence_counter)
    timestamps_message = timestampEnd(builder)

    builder.Finish(timestamps_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


Timestamps = NamedTuple(
    "Timestamps",
    (("name", str), ("timestamps", np.ndarray), ("sequence_counter", int)),
)


def deserialise_tdct(buffer: Union[bytearray, bytes]) -> Timestamps:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    timestamps = timestamp.GetRootAstimestamp(buffer, 0)
    name = timestamps.Name() if timestamps.Name() else b""

    timestamps_array = timestamps.TimestampsAsNumpy()

    return Timestamps(name.decode(), timestamps_array, timestamps.SequenceCounter())



from datetime import datetime, timezone
from typing import List, NamedTuple, Optional, Union

import flatbuffers
import numpy as np

from streaming_data_types.fbschemas.sample_environment_senv.Location import Location
from streaming_data_types.fbschemas.sample_environment_senv.SampleEnvironmentData import (
    SampleEnvironmentData,
    SampleEnvironmentDataAddChannel,
    SampleEnvironmentDataAddMessageCounter,
    SampleEnvironmentDataAddName,
    SampleEnvironmentDataAddPacketTimestamp,
    SampleEnvironmentDataAddTimeDelta,
    SampleEnvironmentDataAddTimestampLocation,
    SampleEnvironmentDataAddTimestamps,
    SampleEnvironmentDataAddValues,
    SampleEnvironmentDataAddValuesType,
    SampleEnvironmentDataEnd,
    SampleEnvironmentDataStart,
)
from streaming_data_types.fbschemas.sample_environment_senv.ValueUnion import ValueUnion
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"senv"


def serialise_senv(
    name: str,
    channel: int,
    timestamp: datetime,
    sample_ts_delta: int,
    message_counter: int,
    values: Union[np.ndarray, List],
    ts_location: Location = Location.Middle,
    value_timestamps: Union[np.ndarray, List, None] = None,
) -> bytes:
    builder = flatbuffers.Builder(1024)

    if value_timestamps is not None:
        used_timestamps = np.atleast_1d(np.asarray(value_timestamps)).astype(np.uint64)
        timestamps_offset = builder.CreateNumpyVector(used_timestamps)

    numpy_type_map = {
        np.dtype("int8"): ValueUnion.Int8Array,
        np.dtype("uint8"): ValueUnion.UInt8Array,
        np.dtype("int16"): ValueUnion.Int16Array,
        np.dtype("uint16"): ValueUnion.UInt16Array,
        np.dtype("int32"): ValueUnion.Int32Array,
        np.dtype("uint32"): ValueUnion.UInt32Array,
        np.dtype("int64"): ValueUnion.Int64Array,
        np.dtype("uint64"): ValueUnion.UInt64Array,
    }

    temp_values = np.atleast_1d(np.asarray(values))

    value_array_offset = builder.CreateNumpyVector(temp_values)

    # Some flatbuffer fu in order to avoid >200 lines of code
    builder.StartObject(1)
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(value_array_offset), 0
    )
    value_offset = builder.EndObject()

    name_offset = builder.CreateString(name)

    SampleEnvironmentDataStart(builder)
    SampleEnvironmentDataAddName(builder, name_offset)
    SampleEnvironmentDataAddTimeDelta(builder, sample_ts_delta)
    SampleEnvironmentDataAddTimestampLocation(builder, ts_location)
    SampleEnvironmentDataAddMessageCounter(builder, message_counter)
    SampleEnvironmentDataAddChannel(builder, channel)
    SampleEnvironmentDataAddPacketTimestamp(builder, int(timestamp.timestamp() * 1e9))
    SampleEnvironmentDataAddValues(builder, value_offset)
    SampleEnvironmentDataAddValuesType(builder, numpy_type_map[temp_values.dtype])
    if value_timestamps is not None:
        SampleEnvironmentDataAddTimestamps(builder, timestamps_offset)

    SE_Message = SampleEnvironmentDataEnd(builder)

    builder.Finish(SE_Message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


Response = NamedTuple(
    "SampleEnvironmentData",
    (
        ("name", str),
        ("channel", int),
        ("timestamp", datetime),
        ("sample_ts_delta", int),
        ("ts_location", Location),
        ("message_counter", int),
        ("values", np.ndarray),
        ("value_ts", Optional[np.ndarray]),
    ),
)


def deserialise_senv(buffer: Union[bytearray, bytes]) -> Response:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    SE_data = SampleEnvironmentData.GetRootAsSampleEnvironmentData(buffer, 0)

    max_time = datetime(
        year=9000, month=1, day=1, hour=0, minute=0, second=0
    ).timestamp()
    used_timestamp = SE_data.PacketTimestamp() / 1e9
    if used_timestamp > max_time:
        used_timestamp = max_time

    value_timestamps = None
    if not SE_data.TimestampsIsNone():
        value_timestamps = SE_data.TimestampsAsNumpy()

    from flatbuffers.number_types import (
        Int8Flags,
        Int16Flags,
        Int32Flags,
        Int64Flags,
        Uint8Flags,
        Uint16Flags,
        Uint32Flags,
        Uint64Flags,
    )

    flag_map = {
        ValueUnion.Int8Array: Int8Flags,
        ValueUnion.UInt8Array: Uint8Flags,
        ValueUnion.Int16Array: Int16Flags,
        ValueUnion.UInt16Array: Uint16Flags,
        ValueUnion.Int32Array: Int32Flags,
        ValueUnion.UInt32Array: Uint32Flags,
        ValueUnion.Int64Array: Int64Flags,
        ValueUnion.UInt64Array: Uint64Flags,
    }

    # Some flatbuffers fu in order to avoid >200 lines of code
    value_offset = SE_data.Values()
    value_type = SE_data.ValuesType()
    values = value_offset.GetVectorAsNumpy(flag_map[value_type], 4)

    return Response(
        name=SE_data.Name().decode(),
        channel=SE_data.Channel(),
        timestamp=datetime.fromtimestamp(used_timestamp, tz=timezone.utc),
        sample_ts_delta=SE_data.TimeDelta(),
        ts_location=SE_data.TimestampLocation(),
        message_counter=SE_data.MessageCounter(),
        values=values,
        value_ts=value_timestamps,
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers


class timestamp(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAstimestamp(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = timestamp()
        x.Init(buf, n + offset)
        return x

    # timestamp
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # timestamp
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # timestamp
    def Timestamps(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # timestamp
    def TimestampsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # timestamp
    def TimestampsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # timestamp
    def SequenceCounter(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0


def timestampStart(builder):
    builder.StartObject(3)


def timestampAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0
    )


def timestampAddTimestamps(builder, timestamps):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(timestamps), 0
    )


def timestampStartTimestampsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def timestampAddSequenceCounter(builder, sequenceCounter):
    builder.PrependUint64Slot(2, sequenceCounter, 0)


def timestampEnd(builder):
    return builder.EndObject()



from datetime import datetime, timezone
from typing import NamedTuple, Union

import flatbuffers

import streaming_data_types.fbschemas.action_response_answ.ActionResponse as ActionResponse
from streaming_data_types.fbschemas.action_response_answ.ActionOutcome import (
    ActionOutcome,
)
from streaming_data_types.fbschemas.action_response_answ.ActionType import ActionType
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"answ"


def serialise_answ(
    service_id: str,
    job_id: str,
    command_id: str,
    action: ActionType,
    outcome: ActionOutcome,
    message: str,
    status_code: int,
    stop_time: datetime,
) -> bytes:
    builder = flatbuffers.Builder(500)
    builder.ForceDefaults(True)
    service_id_offset = builder.CreateString(service_id)
    job_id_offset = builder.CreateString(job_id)
    message_offset = builder.CreateString(message)
    command_id_offset = builder.CreateString(command_id)

    ActionResponse.ActionResponseStart(builder)
    ActionResponse.ActionResponseAddServiceId(builder, service_id_offset)
    ActionResponse.ActionResponseAddJobId(builder, job_id_offset)
    ActionResponse.ActionResponseAddAction(builder, action)
    ActionResponse.ActionResponseAddOutcome(builder, outcome)
    ActionResponse.ActionResponseAddMessage(builder, message_offset)
    ActionResponse.ActionResponseAddCommandId(builder, command_id_offset)
    ActionResponse.ActionResponseAddStatusCode(builder, status_code)
    ActionResponse.ActionResponseAddStopTime(builder, int(stop_time.timestamp() * 1000))

    out_message = ActionResponse.ActionResponseEnd(builder)
    builder.Finish(out_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


Response = NamedTuple(
    "Response",
    (
        ("service_id", str),
        ("job_id", str),
        ("command_id", str),
        ("action", ActionType),
        ("outcome", ActionOutcome),
        ("message", str),
        ("status_code", int),
        ("stop_time", datetime),
    ),
)


def deserialise_answ(buffer: Union[bytearray, bytes]):
    check_schema_identifier(buffer, FILE_IDENTIFIER)
    answ_message = ActionResponse.ActionResponse.GetRootAsActionResponse(buffer, 0)
    max_time = datetime(
        year=9000, month=1, day=1, hour=0, minute=0, second=0
    ).timestamp()
    used_timestamp = answ_message.StopTime() / 1000
    if used_timestamp > max_time:
        used_timestamp = max_time
    return Response(
        service_id=answ_message.ServiceId().decode("utf-8"),
        job_id=answ_message.JobId().decode("utf-8"),
        command_id=answ_message.CommandId().decode("utf-8"),
        action=answ_message.Action(),
        outcome=answ_message.Outcome(),
        message=answ_message.Message().decode("utf-8"),
        status_code=answ_message.StatusCode(),
        stop_time=datetime.fromtimestamp(used_timestamp, tz=timezone.utc),
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class Value(object):
    NONE = 0
    Int8 = 1
    UInt8 = 2
    Int16 = 3
    UInt16 = 4
    Int32 = 5
    UInt32 = 6
    Int64 = 7
    UInt64 = 8
    Float32 = 9
    Float64 = 10
    ArrayInt8 = 11
    ArrayUInt8 = 12
    ArrayInt16 = 13
    ArrayUInt16 = 14
    ArrayInt32 = 15
    ArrayUInt32 = 16
    ArrayInt64 = 17
    ArrayUInt64 = 18
    ArrayFloat32 = 19
    ArrayFloat64 = 20



import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.run_start_pl72 import (
    DetectorSpectrumMap,
    deserialise_pl72,
    serialise_pl72,
)


class TestSerialisationPl72:
    original_entry = {
        "job_id": "some_key",
        "filename": "test_file.nxs",
        "start_time": 567890,
        "stop_time": 578214,
        "run_name": "test_run",
        "nexus_structure": "{}",
        "service_id": "filewriter1",
        "instrument_name": "LOKI",
        "broker": "localhost:9092",
        "metadata": "{3:1}",
        "detector_spectrum_map": DetectorSpectrumMap(
            np.array([4, 5, 6]), np.array([0, 1, 2]), 3
        ),
        "control_topic": "some_topic_name",
    }

    def test_serialises_and_deserialises_pl72_message_correctly(self):
        buf = serialise_pl72(**self.original_entry)
        deserialised_tuple = deserialise_pl72(buf)

        assert deserialised_tuple.job_id == self.original_entry["job_id"]
        assert deserialised_tuple.filename == self.original_entry["filename"]
        assert deserialised_tuple.start_time == self.original_entry["start_time"]
        assert deserialised_tuple.stop_time == self.original_entry["stop_time"]
        assert deserialised_tuple.run_name == self.original_entry["run_name"]
        assert (
            deserialised_tuple.nexus_structure == self.original_entry["nexus_structure"]
        )
        assert deserialised_tuple.service_id == self.original_entry["service_id"]
        assert (
            deserialised_tuple.instrument_name == self.original_entry["instrument_name"]
        )
        assert deserialised_tuple.broker == self.original_entry["broker"]
        assert deserialised_tuple.metadata == self.original_entry["metadata"]

        assert (
            deserialised_tuple.detector_spectrum_map.n_spectra
            == self.original_entry["detector_spectrum_map"].n_spectra
        )
        assert np.array_equal(
            deserialised_tuple.detector_spectrum_map.spectrum_numbers,
            self.original_entry["detector_spectrum_map"].spectrum_numbers,
        )
        assert np.array_equal(
            deserialised_tuple.detector_spectrum_map.detector_ids,
            self.original_entry["detector_spectrum_map"].detector_ids,
        )
        assert deserialised_tuple.control_topic == self.original_entry["control_topic"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        buf = serialise_pl72(**self.original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_pl72(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "pl72" in SERIALISERS
        assert "pl72" in DESERIALISERS



from streaming_data_types.exceptions import ShortBufferException, WrongSchemaException


def get_schema(buffer) -> str:
    """
    Extract the schema code embedded in the buffer

    :param buffer: The raw buffer of the FlatBuffers message.
    :return: The schema identifier
    """
    if len(buffer) < 8:
        raise ShortBufferException("Could not retrieve schema as buffer too short")
    return buffer[4:8].decode("utf-8")


def check_schema_identifier(buffer, expected_identifer: bytes):
    """
    Check the schema code embedded in the buffer matches an expected identifier

    :param buffer: The raw buffer of the FlatBuffers message
    :param expected_identifer: The expected flatbuffer identifier
    """
    if get_schema(buffer) != expected_identifer.decode():
        raise WrongSchemaException(
            f"Incorrect schema: expected {expected_identifer} but got {get_schema(buffer)}"
        )



import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.finished_writing_wrdn import deserialise_wrdn, serialise_wrdn


class TestEncoder(object):
    def test_serialise_and_deserialise_wrdn_message(self):
        """
        Round-trip to check what we serialise is what we get back.
        """

        original_entry = {
            "service_id": "some_service_id_1234",
            "job_id": "some_job_id_abcdef",
            "error_encountered": True,
            "file_name": "somefile.nxs",
            "metadata": '{"hello":4}',
            "message": "some random error message",
        }

        buf = serialise_wrdn(**original_entry)
        entry = deserialise_wrdn(buf)

        assert entry.service_id == original_entry["service_id"]
        assert entry.job_id == original_entry["job_id"]
        assert entry.error_encountered == original_entry["error_encountered"]
        assert entry.file_name == original_entry["file_name"]
        assert entry.metadata == original_entry["metadata"]
        assert entry.message == original_entry["message"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {
            "service_id": "some_service_id_1234",
            "job_id": "some_job_id_abcdef",
            "error_encountered": True,
            "file_name": "somefile.nxs",
            "metadata": '{"hello":4}',
            "message": "some random error message",
        }

        buf = serialise_wrdn(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_wrdn(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "wrdn" in SERIALISERS
        assert "wrdn" in DESERIALISERS



import time
from collections import namedtuple
from datetime import datetime
from typing import NamedTuple, Optional, Union

import flatbuffers
import numpy as np

from streaming_data_types.fbschemas.run_start_pl72 import (
    RunStart,
    SpectraDetectorMapping,
)
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"pl72"


DetectorSpectrumMap = namedtuple(
    "DetectorSpectrumMap",
    (
        "spectrum_numbers",  # numpy ndarray of int
        "detector_ids",  # numpy ndarray of int
        "n_spectra",  # int
    ),
)


def serialise_pl72(
    job_id: str,
    filename: str,
    start_time: Union[int, datetime, None] = None,
    stop_time: Union[int, datetime, None] = None,
    run_name: str = "test_run",
    nexus_structure: str = "{}",
    service_id: str = "",
    instrument_name: str = "TEST",
    broker: str = "localhost:9092",
    metadata: str = "{}",
    detector_spectrum_map: Optional[DetectorSpectrumMap] = None,
    control_topic: str = "",
) -> bytes:
    builder = flatbuffers.Builder(512)
    builder.ForceDefaults(True)

    if type(start_time) is datetime:
        start_time = int(start_time.timestamp() * 1000)
    elif start_time is None:
        start_time = int(time.time() * 1000)
    if service_id is None:
        service_id = ""
    if type(stop_time) is datetime:
        stop_time = int(stop_time.timestamp() * 1000)
    elif stop_time is None:
        stop_time = 0

    service_id_offset = builder.CreateString(service_id)
    broker_offset = builder.CreateString(broker)
    job_id_offset = builder.CreateString(job_id)
    nexus_structure_offset = builder.CreateString(nexus_structure)
    instrument_name_offset = builder.CreateString(instrument_name)
    run_name_offset = builder.CreateString(run_name)
    filename_offset = builder.CreateString(filename)
    metadata_offset = builder.CreateString(metadata)
    control_topic_offset = builder.CreateString(control_topic)

    # Build detector-spectrum map
    if detector_spectrum_map is not None:
        spectrum_map_offset = builder.CreateNumpyVector(
            np.asarray(detector_spectrum_map.spectrum_numbers).astype(np.int32)
        )
        det_id_map_offset = builder.CreateNumpyVector(
            np.asarray(detector_spectrum_map.detector_ids).astype(np.int32)
        )
        SpectraDetectorMapping.SpectraDetectorMappingStart(builder)
        SpectraDetectorMapping.SpectraDetectorMappingAddSpectrum(
            builder, spectrum_map_offset
        )
        SpectraDetectorMapping.SpectraDetectorMappingAddDetectorId(
            builder, det_id_map_offset
        )
        SpectraDetectorMapping.SpectraDetectorMappingAddNSpectra(
            builder, detector_spectrum_map.n_spectra
        )
        detector_spectrum_map_offset = SpectraDetectorMapping.SpectraDetectorMappingEnd(
            builder
        )

    # Build the actual buffer
    RunStart.RunStartStart(builder)
    RunStart.RunStartAddServiceId(builder, service_id_offset)
    RunStart.RunStartAddBroker(builder, broker_offset)
    RunStart.RunStartAddJobId(builder, job_id_offset)
    RunStart.RunStartAddNexusStructure(builder, nexus_structure_offset)
    RunStart.RunStartAddInstrumentName(builder, instrument_name_offset)
    RunStart.RunStartAddRunName(builder, run_name_offset)
    RunStart.RunStartAddStopTime(builder, stop_time)
    RunStart.RunStartAddStartTime(builder, start_time)
    RunStart.RunStartAddFilename(builder, filename_offset)
    RunStart.RunStartAddNPeriods(builder, 1)
    RunStart.RunStartAddMetadata(builder, metadata_offset)
    if detector_spectrum_map is not None:
        RunStart.RunStartAddDetectorSpectrumMap(builder, detector_spectrum_map_offset)
    RunStart.RunStartAddControlTopic(builder, control_topic_offset)

    run_start_message = RunStart.RunStartEnd(builder)

    builder.Finish(run_start_message, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


class RunStartInfo(NamedTuple):
    job_id: str
    filename: str
    start_time: int
    stop_time: int
    nexus_structure: str
    service_id: str
    broker: str
    run_name: str = ""
    instrument_name: str = ""
    metadata: str = ""
    detector_spectrum_map: Optional[DetectorSpectrumMap] = None
    control_topic: str = ""


def deserialise_pl72(buffer: Union[bytearray, bytes]) -> RunStartInfo:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    run_start = RunStart.RunStart.GetRootAsRunStart(buffer, 0)
    service_id = run_start.ServiceId() if run_start.ServiceId() else b""
    broker = run_start.Broker() if run_start.Broker() else b""
    job_id = run_start.JobId() if run_start.JobId() else b""
    filename = run_start.Filename() if run_start.Filename() else b""
    nexus_structure = run_start.NexusStructure() if run_start.NexusStructure() else b""
    instrument_name = run_start.InstrumentName() if run_start.InstrumentName() else b""
    run_name = run_start.RunName() if run_start.RunName() else b""
    metadata = run_start.Metadata() if run_start.Metadata() else b""
    control_topic = run_start.ControlTopic() if run_start.ControlTopic() else b""

    detector_spectrum_map = None
    det_spec_map_buf = run_start.DetectorSpectrumMap()
    if det_spec_map_buf is not None:
        detector_spectrum_map = DetectorSpectrumMap(
            det_spec_map_buf.SpectrumAsNumpy(),
            det_spec_map_buf.DetectorIdAsNumpy(),
            det_spec_map_buf.NSpectra(),
        )

    return RunStartInfo(
        job_id=job_id.decode(),
        filename=filename.decode(),
        start_time=run_start.StartTime(),
        stop_time=run_start.StopTime(),
        run_name=run_name.decode(),
        nexus_structure=nexus_structure.decode(),
        service_id=service_id.decode(),
        instrument_name=instrument_name.decode(),
        broker=broker.decode(),
        metadata=metadata.decode(),
        detector_spectrum_map=detector_spectrum_map,
        control_topic=control_topic.decode(),
    )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Event43Message(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsEvent43Message(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Event43Message()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def Event43MessageBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x65\x76\x34\x33", size_prefixed=size_prefixed
        )

    # Event43Message
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Event43Message
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Event43Message
    def MessageId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Event43Message
    def PulseTime(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # Event43Message
    def PulseTimeAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # Event43Message
    def PulseTimeLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event43Message
    def PulseTimeIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Event43Message
    def PulseIndex(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # Event43Message
    def PulseIndexAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # Event43Message
    def PulseIndexLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event43Message
    def PulseIndexIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Event43Message
    def TimeOfFlight(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # Event43Message
    def TimeOfFlightAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # Event43Message
    def TimeOfFlightLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event43Message
    def TimeOfFlightIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # Event43Message
    def DetectorId(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4),
            )
        return 0

    # Event43Message
    def DetectorIdAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # Event43Message
    def DetectorIdLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Event43Message
    def DetectorIdIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0


def Event43MessageStart(builder):
    builder.StartObject(6)


def Event43MessageAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def Event43MessageAddMessageId(builder, messageId):
    builder.PrependUint64Slot(1, messageId, 0)


def Event43MessageAddPulseTime(builder, pulseTime):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(pulseTime), 0
    )


def Event43MessageStartPulseTimeVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def Event43MessageAddPulseIndex(builder, pulseIndex):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(pulseIndex), 0
    )


def Event43MessageStartPulseIndexVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def Event43MessageAddTimeOfFlight(builder, timeOfFlight):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(timeOfFlight), 0
    )


def Event43MessageStartTimeOfFlightVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def Event43MessageAddDetectorId(builder, detectorId):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(detectorId), 0
    )


def Event43MessageStartDetectorIdVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def Event43MessageEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:


class DType(object):
    int8 = 0
    uint8 = 1
    int16 = 2
    uint16 = 3
    int32 = 4
    uint32 = 5
    int64 = 6
    uint64 = 7
    float32 = 8
    float64 = 9
    c_string = 10



# automatically generated by the FlatBuffers compiler, do not modify

# namespace:

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ADArray(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsADArray(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ADArray()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def ADArrayBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x41\x44\x41\x72", size_prefixed=size_prefixed
        )

    # ADArray
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ADArray
    def SourceName(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ADArray
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ADArray
    def Timestamp(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # ADArray
    def Dimensions(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 8),
            )
        return 0

    # ADArray
    def DimensionsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint64Flags, o)
        return 0

    # ADArray
    def DimensionsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ADArray
    def DimensionsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # ADArray
    def DataType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # ADArray
    def Data(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # ADArray
    def DataAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # ADArray
    def DataLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ADArray
    def DataIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # ADArray
    def Attributes(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .Attribute import Attribute

            obj = Attribute()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ADArray
    def AttributesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # ADArray
    def AttributesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        return o == 0


def ADArrayStart(builder):
    builder.StartObject(7)


def ADArrayAddSourceName(builder, sourceName):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(sourceName), 0
    )


def ADArrayAddId(builder, id):
    builder.PrependInt32Slot(1, id, 0)


def ADArrayAddTimestamp(builder, timestamp):
    builder.PrependUint64Slot(2, timestamp, 0)


def ADArrayAddDimensions(builder, dimensions):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(dimensions), 0
    )


def ADArrayStartDimensionsVector(builder, numElems):
    return builder.StartVector(8, numElems, 8)


def ADArrayAddDataType(builder, dataType):
    builder.PrependInt8Slot(4, dataType, 0)


def ADArrayAddData(builder, data):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(data), 0
    )


def ADArrayStartDataVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def ADArrayAddAttributes(builder, attributes):
    builder.PrependUOffsetTRelativeSlot(
        6, flatbuffers.number_types.UOffsetTFlags.py_type(attributes), 0
    )


def ADArrayStartAttributesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def ADArrayEnd(builder):
    return builder.EndObject()



from typing import Optional, Union, NamedTuple
from datetime import datetime, timezone

import flatbuffers

from streaming_data_types.fbschemas.epics_conn_status_pvCn.ConnectionInfo import (
    ConnectionInfo,
)
from streaming_data_types.fbschemas.epics_conn_status_pvCn import EpicsPVConnectionInfo
from streaming_data_types.utils import check_schema_identifier

FILE_IDENTIFIER = b"pvCn"


def serialise_pvCn(
    timestamp: datetime,
    status: ConnectionInfo,
    source_name: str,
    service_id: Optional[str] = None,
) -> bytes:
    builder = flatbuffers.Builder(136)
    builder.ForceDefaults(True)

    if service_id is not None:
        service_id_offset = builder.CreateString(service_id)
    source_name_offset = builder.CreateString(source_name)

    EpicsPVConnectionInfo.EpicsPVConnectionInfoStart(builder)
    if service_id is not None:
        EpicsPVConnectionInfo.EpicsPVConnectionInfoAddServiceId(
            builder, service_id_offset
        )
    EpicsPVConnectionInfo.EpicsPVConnectionInfoAddSourceName(
        builder, source_name_offset
    )
    EpicsPVConnectionInfo.EpicsPVConnectionInfoAddStatus(builder, status)
    EpicsPVConnectionInfo.EpicsPVConnectionInfoAddTimestamp(
        builder, int(timestamp.timestamp() * 1e6)
    )

    end = EpicsPVConnectionInfo.EpicsPVConnectionInfoEnd(builder)
    builder.Finish(end, file_identifier=FILE_IDENTIFIER)
    return bytes(builder.Output())


EpicsPVConnection = NamedTuple(
    "EpicsPVConnection",
    (
        ("timestamp", datetime),
        ("status", ConnectionInfo),
        ("source_name", str),
        ("service_id", Optional[str]),
    ),
)


def deserialise_pvCn(buffer: Union[bytearray, bytes]) -> EpicsPVConnection:
    check_schema_identifier(buffer, FILE_IDENTIFIER)

    epics_connection = EpicsPVConnectionInfo.EpicsPVConnectionInfo.GetRootAs(buffer, 0)

    source_name = (
        epics_connection.SourceName() if epics_connection.SourceName() else b""
    )
    service_id = (
        epics_connection.ServiceId().decode() if epics_connection.ServiceId() else None
    )

    return EpicsPVConnection(
        timestamp=datetime.fromtimestamp(
            epics_connection.Timestamp() / 1e6, tz=timezone.utc
        ),
        status=epics_connection.Status(),
        source_name=source_name.decode(),
        service_id=service_id,
    )



from datetime import datetime, timezone

import numpy as np
import pytest

from streaming_data_types import DESERIALISERS, SERIALISERS
from streaming_data_types.area_detector_ADAr import (
    Attribute,
    deserialise_ADAr,
    serialise_ADAr,
)
from streaming_data_types.exceptions import WrongSchemaException


class TestSerialisationNDAr:
    def test_serialises_and_deserialises_ADAr_int_array(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "source_name": "some source name",
            "unique_id": 754,
            "data": np.array([[1, 2, 3], [3, 4, 5]], dtype=np.uint64),
            "timestamp": datetime.now(tz=timezone.utc),
            "attributes": [
                Attribute("name1", "desc1", "src1", "value"),
                Attribute("name2", "desc2", "src2", 11),
                Attribute("name3", "desc3", "src3", 3.14),
                Attribute("name4", "desc4", "src4", np.linspace(0, 10)),
            ],
        }

        buf = serialise_ADAr(**original_entry)
        entry = deserialise_ADAr(buf)

        assert entry.unique_id == original_entry["unique_id"]
        assert entry.source_name == original_entry["source_name"]
        assert entry.timestamp == original_entry["timestamp"]
        assert np.array_equal(entry.data, original_entry["data"])
        assert entry.data.dtype == original_entry["data"].dtype
        assert len(entry.attributes) == len(original_entry["attributes"])
        for i in range(len(entry.attributes)):
            assert entry.attributes[i] == original_entry["attributes"][i]

    def test_serialises_and_deserialises_ADAr_float_array(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "source_name": "some other source name",
            "unique_id": 789679,
            "data": np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]], dtype=np.float32),
            "timestamp": datetime(
                year=1992,
                month=8,
                day=11,
                hour=3,
                minute=34,
                second=57,
                tzinfo=timezone.utc,
            ),
        }

        buf = serialise_ADAr(**original_entry)
        entry = deserialise_ADAr(buf)

        assert entry.unique_id == original_entry["unique_id"]
        assert entry.source_name == original_entry["source_name"]
        assert entry.timestamp == original_entry["timestamp"]
        assert np.array_equal(entry.data, original_entry["data"])
        assert entry.data.dtype == original_entry["data"].dtype

    def test_serialises_and_deserialises_ADAr_string(self):
        """
        Round-trip to check what we serialise is what we get back.
        """
        original_entry = {
            "source_name": "some source name",
            "unique_id": 754,
            "data": "hi, this is a string",
            "timestamp": datetime.now(tz=timezone.utc),
        }

        buf = serialise_ADAr(**original_entry)
        entry = deserialise_ADAr(buf)

        assert entry.unique_id == original_entry["unique_id"]
        assert entry.source_name == original_entry["source_name"]
        assert entry.timestamp == original_entry["timestamp"]
        assert entry.data == original_entry["data"]

    def test_if_buffer_has_wrong_id_then_throws(self):
        original_entry = {
            "source_name": "some source name",
            "unique_id": 754,
            "data": np.array([[1, 2, 3], [3, 4, 5]], dtype=np.uint64),
            "timestamp": datetime.now(),
        }

        buf = serialise_ADAr(**original_entry)

        # Manually hack the id
        buf = bytearray(buf)
        buf[4:8] = b"1234"

        with pytest.raises(WrongSchemaException):
            deserialise_ADAr(buf)

    def test_schema_type_is_in_global_serialisers_list(self):
        assert "ADAr" in SERIALISERS
        assert "ADAr" in DESERIALISERS
